import logging
from typing import Callable, List, Optional

import pandas as pd

from oogeso import dto
from oogeso.core import devices

logger = logging.getLogger(__name__)


def get_device_from_model_name(model_name: str) -> Callable:
    map_device_name_to_class = {
        "powersource": devices.Powersource,
        "powersink": devices.PowerSink,
        "storageel": devices.StorageEl,
        "compressorel": devices.CompressorEl,
        "compressorgas": devices.CompressorGas,
        "electrolyser": devices.Electrolyser,
        "fuelcell": devices.FuelCell,
        "gasheater": devices.GasHeater,
        "gasturbine": devices.GasTurbine,
        "heatpump": devices.HeatPump,
        "pumpoil": devices.PumpOil,
        "pumpwater": devices.PumpWater,
        "separastor": devices.Separator,
        "separator2": devices.Separator2,
        "sinkel": devices.SinkEl,
        "sinkheat": devices.SinkHeat,
        "sinkgas": devices.SinkGas,
        "sinkoil": devices.SinkOil,
        "sinkwater": devices.SinkWater,
        "sourceel": devices.SourceEl,
        "sourcegas": devices.SourceGas,
        "sourceoil": devices.SourceOil,
        "sourcewater": devices.SourceWater,
        "storagehydrogen": devices.StorageHydrogen,
        "wellgaslift": devices.WellGasLift,
        "wellproduction": devices.WellProduction,
    }
    if model_name in map_device_name_to_class:
        return map_device_name_to_class[model_name]
    else:
        raise NotImplementedError(f"Device {model_name} has not been implemented.")


def get_class_from_dto(class_str: str) -> Callable:
    """
    Search dto module for a callable that matches the signature given as class str

    Fixme: Replace this (de-)serializer with a proper solution.
    """
    if class_str in dto.__dict__.keys():
        return dto.__dict__[class_str]
    elif class_str.lower() in [x.lower() for x in dto.__dict__.keys()]:
        return [v for k, v in dto.__dict__.items() if k.lower() == class_str.lower()][0]
    elif class_str.lower().replace("_", "") in [x.lower() for x in dto.__dict__.keys()]:
        return [v for k, v in dto.__dict__.items() if k.lower() == class_str.lower().replace("_", "")][0]
    else:
        raise NotImplementedError(f"Model {class_str} has not been implemented.")


def create_time_series_data(
    df_forecast: pd.DataFrame,
    df_nowcast: pd.DataFrame,
    time_start: Optional[str],
    time_end: Optional[str],
    timestep_minutes: int,
    resample_method: str = "linear",
) -> List[dto.TimeSeriesData]:
    """Rearrange and resample pandas timeseries to Oogeso data transfer object

    The input dataframes should have a datetime index
    """

    # Compine forecast and nowcast timeseries into a single dataframe
    df_orig = pd.concat({"forecast": df_forecast, "nowcast": df_nowcast}, axis=1)
    # 1 resample
    resampled = df_orig.resample(
        rule="{}T".format(
            timestep_minutes,
        ),
        closed="left",
        label="left",
    )
    df_new = resampled.mean()

    # If up-sampling, NaN values filed by linear interpolation (or other method)
    # If down-sampling, this does nothing
    df_new = df_new.interpolate(method=resample_method)

    # Select the times of interest
    if time_start is not None:
        mask = df_new.index >= time_start
        df_new = df_new[mask]
    if time_end is not None:
        mask = df_new.index <= time_end
        df_new = df_new[mask]
    if df_new.isna().any().any():
        logger.warning("Profiles contain NA")

    # now, create Oogeso TimeSereriesData object
    profiles = []
    for col in df_new.columns:
        if col[0] == "forecast":
            curve = col[1]
            list_data = list(df_new[("forecast", curve)])
            list_data_nowcast = None
            if ("nowcast", curve) in df_new.columns:
                list_data_nowcast = list(df_new[("nowcast", curve)])
            new_ts = dto.TimeSeriesData(id=curve, data=list_data, data_nowcast=list_data_nowcast)
            profiles.append(new_ts)
        elif col[0] == "nowcast":
            if ("forecast", curve) not in df_new.columns:  # Fixme: curve potentially referenced before assignment.
                logger.warning("Nowcast but no forecast profile for {}".format(curve))
    return profiles



import logging
from typing import Tuple, Union

import numpy as np
import pyomo.environ as pyo
import scipy

from oogeso import dto
from oogeso.core.networks.edge import Edge
from oogeso.core.networks.network import Network

logger = logging.getLogger(__name__)

GRAVITY_ACCELERATION_CONSTANT = 9.8  # m/s^2


class Fluid(Network):
    def define_constraints(self, pyomo_model: pyo.Model):
        super().define_constraints(pyomo_model)

        # additional
        if self.carrier_data.pressure_method in ["weymouth", "darcy-weissbach"]:
            edgelist = [e for e in self.edges.keys()]
            if edgelist:
                # non-empty list of edges
                constr_flow = pyo.Constraint(
                    edgelist,
                    pyomo_model.setHorizon,
                    rule=self._rulePipelineFlow,
                )
                setattr(
                    pyomo_model,
                    "constrE_{}_{}".format(self.carrier_id, "flow"),
                    constr_flow,
                )

    def _rulePipelineFlow(self, model: pyo.Model, edge, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Pipeline flow vs pressure drop"""
        # edge = self.id
        edge_obj = self.edges[edge]
        edge_data = edge_obj.edge_data
        carrier = edge_data.carrier
        n_from = edge_data.node_from
        n_to = edge_data.node_to
        print_log = True if t == 0 else False

        p1 = model.varPressure[(n_from, carrier, "out", t)]
        p2 = model.varPressure[(n_to, carrier, "in", t)]
        Q = model.varEdgeFlow[edge, t]
        if edge_data.num_pipes is not None:
            num_pipes = edge_data.num_pipes
            if print_log:
                logger.debug("{},{}: {} parallel pipes".format(edge, t, num_pipes))
            Q = Q / num_pipes
        p2_computed = self.compute_edge_pressuredrop(edge_obj, p1=p1, Q=Q, linear=True, print_log=print_log)
        return p2 == p2_computed

    def _compute_exps_and_k(self, edge_data: dto.EdgeFluidData, carrier_data: dto.CarrierData):
        """Derive exp_s and k parameters for Weymouth equation"""
        # gas pipeline parameters - derive k and exp(s) parameters:
        ga = carrier_data

        # if 'temperature_K' in model.paramEdge[edge]:
        temp = edge_data.temperature_K
        height_difference = edge_data.height_m
        length = edge_data.length_km
        diameter = edge_data.diameter_mm
        s = 0.0684 * (ga.G_gravity * height_difference / (temp * ga.Z_compressibility))
        if s > 0:
            # height difference - use equivalent length
            sfactor = (np.exp(s) - 1) / s
            length = length * sfactor

        k = (
            4.3328e-8
            * ga.Tb_basetemp_K
            / ga.Pb_basepressure_MPa
            * (ga.G_gravity * temp * length * ga.Z_compressibility) ** (-1 / 2)
            * diameter ** (8 / 3)
        )
        exp_s = np.exp(s)
        return exp_s, k

    def compute_edge_pressuredrop(self, edge: "Edge", p1, Q, method=None, linear=False, print_log=True):
        """Compute pressure drop in pipe

        parameters
        ----------
        edge : edge object
        p1 : float
            pipe inlet pressure (MPa)
        Q : float
            flow rate (Sm3/s)
        method : string
            None, weymouth, darcy-weissbach
        linear : boolean
            whether to use linear model or not

        Returns
        -------
        p2 : float
            pipe outlet pressure (MPa)"""

        # edge = self.id
        edge_data = edge.edge_data
        edge_id = edge_data.id
        method = None
        carrier = edge_data.carrier
        carrier_data = self.carrier_data
        if carrier_data.pressure_method is not None:
            method = carrier_data.pressure_method

        n_from = edge_data.node_from
        n_to = edge_data.node_to
        p0_from = edge.node_from.get_pressure_nominal(carrier, "out")
        p0_to = edge.node_to.get_pressure_nominal(carrier, "in")
        if (p0_from is not None) and (p0_to is not None):
            if linear & (p0_from == p0_to):
                method = None
                if print_log:
                    logger.debug(
                        ("{}-{}: Pipe without pressure drop" " ({} / {} MPa)").format(n_from, n_to, p0_from, p0_to)
                    )
        elif linear:
            # linear equations, but nominal values not given - assume no drop
            # logger.debug("{}-{}: Aassuming no  pressure drop".format(n_from, n_to))
            method = None
        else:
            # use non-linear equations, no nominal pressure required
            pass

        if method is None:
            # no pressure drop
            p2 = p1
            return p2

        elif method == "weymouth":
            """
            Q = k * sqrt( Pin^2 - e^s Pout^2 ) [Weymouth equation, nonlinear]
                => Pout = sqrt(Pin^2-Q^2/k^2)/e^2
            Q = c * (Pin_0 Pin - e^s Pout0 Pout) [linearised version]
                => Pout = (Pin0 Pin - Q/c)/(e^s Pout0)
            c = k/sqrt(Pin0^2 - e^s Pout0^2)

            REFERENCES:
            1) E Sashi Menon, Gas Pipeline Hydraulics, Taylor & Francis (2005),
            https://doi.org/10.1201/9781420038224
            2) A Tomasgard et al., Optimization  models  for  the  natural  gas
            value  chain, in: Geometric Modelling, Numerical Simulation and
            Optimization. Springer Verlag, New York (2007),
            https://doi.org/10.1007/978-3-540-68783-2_16
            """
            exp_s, k = self._compute_exps_and_k(edge_data, carrier_data=carrier_data)
            if print_log:
                logger.debug("pipe {}: exp_s={}, k={}".format(edge_id, exp_s, k))
            if linear:
                p_from = p1
                #                p_from = model.varPressure[(n_from,carrier,'out',t)]
                #                p_to = model.varPressure[(n_to,carrier,'in',t)]
                X0 = p0_from ** 2 - exp_s * p0_to ** 2
                #                logger.info("edge {}-{}: X0={}, p1={},Q={}"
                #                    .format(n_from,n_to,X0,p1,Q))
                coeff = k * (X0) ** (-1 / 2)
                #                Q_computed = coeff*(p0_from*p_from - exp_s*p0_to*p_to)
                p2 = (p0_from * p_from - Q / coeff) / (exp_s * p0_to)
            else:
                # weymouth eqn (non-linear)
                p2 = 1 / exp_s * (p1 ** 2 - Q ** 2 / k ** 2) ** (1 / 2)
            return p2

        elif method == "darcy-weissbach":
            rho = carrier_data.rho_density
            D = edge_data.diameter_mm / 1000
            L = edge_data.length_km * 1000
            height_difference = edge_data.height_m

            if (carrier_data.viscosity is not None) & (not linear):
                # compute darcy friction factor from flow rate and viscosity (nonlinear)
                mu = carrier_data.viscosity
                Re = 2 * rho * Q / (np.pi * mu * D)
                f = 1 / (0.838 * scipy.special.lambertw(0.629 * Re)) ** 2
                f = f.real
            elif carrier_data.darcy_friction is not None:
                f = carrier_data.darcy_friction
                Re = None
            else:
                raise Exception("Must provide viscosity or darcy_friction for {}".format(carrier))
            (p2, Q0) = darcy_weissbach_p2(
                Q,
                p1 * 1e6,
                f=f,
                rho=rho,
                diameter=D,
                length=L,
                height_difference=height_difference,
                linear=linear,
                p0_from=p0_from * 1e6,
                p0_to=p0_to * 1e6,
            )
            p2 = p2 * 1e-6  # Convert to MPa
            if print_log:
                logger.debug(
                    ("derived pipe ({}) flow rate:" " Q={}, linearQ0={:5.3g}," " friction={:5.3g}").format(
                        edge_id, Q, Q0, f
                    )
                )

            return p2
        else:
            raise Exception("Unknown pressure drop calculation method ({})".format(method))


class Gas(Fluid):
    pass


class Oil(Fluid):
    pass


class Water(Fluid):
    pass


class Wellstream(Fluid):
    pass


def darcy_weissbach_p2(
    Q,
    p1,
    f,
    rho,
    diameter,
    length,
    height_difference=0,
    linear=True,
    p0_from=None,  # Pa
    p0_to=None,  # Pa
) -> Tuple[float, float]:
    """compute outlet pressure from darcy-weissbach equation

    parameters
    ----------
    p1 : float
        pressure at pipe inlet (Pa)
    Q : float
        flow rate (Sm3/s)
    f : float
        friction factor
    rho : float
        fluid density (kg/m3)
    diameter : float
        pipe inner diameter (m)
    length : float
        pipe length (m)
    height_difference : float
        height difference outlet vs inlet (m)

    """
    D = diameter  # m
    L = length  # m
    grav = GRAVITY_ACCELERATION_CONSTANT
    q0 = np.nan
    if linear:
        k2 = (8 * f * rho * L) / (np.pi ** 2 * D ** 5)
        q0_squared = 1 / k2 * ((p0_from - p0_to) - rho * grav * height_difference)
        if q0_squared < 0:
            logging.error(
                "Negative q0^2 (flow direction error) q0^2=%s p1=%s p2=%s rho=%s grav=%s z=%s",
                q0_squared,
                p0_from,
                p0_to,
                rho,
                grav,
                height_difference,
            )
        q0 = np.sqrt(q0_squared)
        p2 = p1 - rho * grav * height_difference - k2 * (2 * q0 * Q - q0 ** 2)
    else:
        # Quadratic in Q
        p2 = p1 - rho * grav * height_difference - 8 * f * rho * L * Q ** 2 / (np.pi ** 2 * D ** 5)
    return (p2, q0)


def darcy_weissbach_Q(p1, p2, f, rho, diameter_mm, length_km, height_difference_m=0):
    """compute flow rate from non-linear darcy-weissbach eqn

    parameters
    ----------
    p1 : float
        pressure at pipe input (MPa)
    p2 : float
        pressure at pipe output (MPa)
    f : float
        friction factor
    rho : float
        fluid density (kg/m3)
    diameter_mm : float
        pipe inner diameter (mm)
    length_km : float
        pipe length (km)
    height_difference_m : float
        height difference output vs input (m)

    """

    grav = 9.98
    L = length_km * 1000
    D = diameter_mm / 1000
    k = 8 * f * rho * L / (np.pi ** 2 * D ** 5)
    Q = np.sqrt(((p1 - p2) * 1e6 - rho * grav * height_difference_m) / k)
    return Q



"""This module contains methods to read and save to file"""

import json
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

import pandas as pd
import yaml

from oogeso import dto
from oogeso.dto.serialisation import DataclassJSONDecoder

logger = logging.getLogger(__name__)


def read_data_from_yaml(filename, profiles=None, profiles_nowcast=None) -> dto.EnergySystemData:
    """Read input data from yaml file"""
    data_dict = None
    with open(filename, "r") as text_file:
        data_dict = yaml.safe_load(text_file)

    if profiles is not None:
        logger.debug("Adding profiles to data in yaml file")
        if "profiles" in data_dict:
            logger.warning("Overiding profile data in yaml file")
        data_dict["profiles"] = profiles
    if profiles_nowcast is not None:
        logger.debug("Adding profiles to data in yaml file")
        if "profiles_nowcast" in data_dict:
            logger.warning("Overiding profile data in yaml file")
        data_dict["profiles_nowcast"] = profiles_nowcast
    json_str = json.dumps(data_dict)
    decoder = DataclassJSONDecoder()
    energy_system = decoder.decode(json_str)
    return energy_system


def read_profiles_from_parquet(filename: Union[Path, str], keys: List[str]) -> Dict[str, pd.DataFrame]:
    """
    Read input data profiles from parquet file

    Reading from the given filename with _<key name><suffix>.
    This is to give legacy support for the hdf5 format where you have keys.

    Fixme: Refactor to avoid this complexity if not needed.
    """
    if not isinstance(filename, Path):
        filename = Path(filename)

    suffix = ""
    if "." in filename.name:
        suffix = filename.suffix

    profiles = {}
    for key in keys:
        profiles[key] = pd.read_parquet(f"{filename.with_suffix('')}_{key}{suffix}")
    return profiles


def save_profiles_to_parquet(filename: Union[Path, str], profiles: Dict[str, pd.DataFrame]) -> None:
    """
    Save profiles to parquet

    Writing to the given filename with _<key name><suffix>.
    This is to give legacy support for the hdf5 format where you have keys.

    Fixme: Refactor to avoid this complexity if not needed.
    """
    if not isinstance(filename, Path):
        filename = Path(filename)

    suffix = ""
    if "." in filename.name:
        suffix = filename.suffix

    for key in profiles:
        profiles[key].to_parquet(f"{filename.with_suffix('')}_{key}{suffix}")
    return


def read_profiles_from_xlsx(
    filename: Path,
    sheet_forecast: str = "profiles",
    sheet_nowcast: str = "profiles_forecast",
    exclude_cols: Optional[List] = None,
) -> Dict[str, pd.DataFrame]:
    """Read input data profiles from XLSX to a dicitonary of pandas dataframes"""
    df_profiles = pd.read_excel(
        filename,
        sheet_name=sheet_forecast,
        index_col="timestep",
        usecols=lambda col: col not in exclude_cols,
    )
    df_profiles_forecast = pd.read_excel(
        filename,
        sheet_name=sheet_nowcast,
        index_col="timestep",
        usecols=lambda col: col not in exclude_cols,
    )
    profiles = {"actual": df_profiles, "forecast": df_profiles_forecast}
    return profiles


def read_profiles_from_csv(
    filename_forecasts,
    filename_nowcasts=None,
    timestamp_col="timestamp",
    dayfirst=True,
    exclude_cols=None,
):
    """Read input data profiles from CSV to a dicitonary of pandas dataframes"""
    if exclude_cols is None:
        exclude_cols = []
    df_profiles_forecast = pd.read_csv(
        filename_forecasts,
        index_col=timestamp_col,
        parse_dates=[timestamp_col],
        dayfirst=dayfirst,
        usecols=lambda col: col not in exclude_cols,
    )
    df_profiles_nowcast = pd.DataFrame()  # empty dataframe
    if filename_nowcasts is not None:
        df_profiles_nowcast = pd.read_csv(
            filename_nowcasts,
            index_col=timestamp_col,
            parse_dates=[timestamp_col],
            dayfirst=dayfirst,
            usecols=lambda col: col not in exclude_cols,
        )

    return {"forecast": df_profiles_forecast, "nowcast": df_profiles_nowcast}



"""This module is for electrical computations
"""
import networkx as nx
import pandas as pd
import scipy

elbase = {"baseMVA": 100, "baseAngle": 1}


def computePowerFlowMatrices(nodes, branches, baseZ=1):
    """
    Compute and return dc power flow matrices B' and DA

    Parameters
    ==========
    nodes : list of nodes (their ids)
    branches : dictionary {key:edge}
            edge is itself a dictionay with fields:
            (node_from, node_to, index, reactance_pu)
    baseZ : float (impedance should already be in pu.)
            base value for impedance

    Returns
    =======
    (coeff_B, coeff_DA) : dictionary of matrix values

    """

    df_branch = pd.DataFrame.from_dict(branches, orient="index")
    susceptance = 1 / df_branch["reactance"] * baseZ
    nodes = list(nodes)
    node_ids = nodes
    edge_ids = []
    edges = []
    for br_id, branch in branches.items():
        b = susceptance[br_id]
        edges.append((branch["node_from"], branch["node_to"], br_id, {"i": br_id, "b": b}))
        edge_ids.append(br_id)

    # MultiDiGraph to allow parallel lines
    G = nx.MultiDiGraph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)
    A_incidence_matrix = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges).T
    # Diagonal matrix
    D = scipy.sparse.diags(-susceptance, offsets=0)
    DA = D * A_incidence_matrix

    # Bf constructed from incidence matrix with branch susceptance
    # used as weight (this is quite fast)
    Bf = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges, weight="b").T
    Bbus = A_incidence_matrix.T * Bf

    coeff_B = dict()
    coeff_DA = dict()
    cx = scipy.sparse.coo_matrix(Bbus)
    for i, j, v in zip(cx.row, cx.col, cx.data):
        coeff_B[(node_ids[i], node_ids[j])] = v

    cx = scipy.sparse.coo_matrix(DA)
    for i, j, v in zip(cx.row, cx.col, cx.data):
        coeff_DA[(edge_ids[i], node_ids[j])] = v

    return coeff_B, coeff_DA


# def _susceptancePu(df_edge,baseOhm=1):
#    '''If impedance is already given in pu, baseOhm should be 1
#    If not, well... baseOhm depends on the voltage level, so need to know
#    the nominal voltage at the bus to convert from ohm to pu.
#    '''
#    #return [-1/self.branch['reactance'][i]*baseOhm
#    #        for i in self.branch.index.tolist()]
#    return 1/df_edge['reactance']*baseOhm



from typing import Dict, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class FuelCell(Device):
    """Fuel cell - hydrogen to electricity."""

    carrier_in = ["hydrogen"]
    carrier_out = ["el", "heat"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceFuelCellData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierHydrogenData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data = self.dev_data
        param_hydrogen = self.carrier_data["hydrogen"]
        energy_value = param_hydrogen.energy_value  # MJ/Sm3
        efficiency = dev_data.eta
        eta_heat = dev_data.eta_heat  # heat recovery efficiency
        if i == 1:
            """hydrogen to el"""
            lhs = pyomo_model.varDeviceFlow[dev, "el", "out", t]  # MW
            rhs = pyomo_model.varDeviceFlow[dev, "hydrogen", "in", t] * energy_value * efficiency
            return lhs == rhs
        elif i == 2:
            """heat output = waste energy * heat recovery factor"""
            lhs = pyomo_model.varDeviceFlow[dev, "heat", "out", t]
            rhs = pyomo_model.varDeviceFlow[dev, "hydrogen", "in", t] * energy_value * (1 - efficiency) * eta_heat
            return lhs == rhs
        else:
            raise ValueError(f"Argument i must be 1 or 2. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        list_to_reconstruct = super().define_constraints(pyomo_model)
        constr = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 2), rule=self._rules)
        # add constraint to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "out", t]



import logging
from dataclasses import asdict
from typing import Dict, Optional, Tuple, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device
from oogeso.core.networks import electricalsystem as el_calc
from oogeso.core.networks.network import Network

logger = logging.getLogger(__name__)


class El(Network):
    def __init__(
        self,
        carrier_data: dto.CarrierElData,
        edges: Dict[str, dto.EdgeElData],
    ):
        super().__init__(carrier_data=carrier_data, edges=edges)
        self.carrier_data = carrier_data
        self.edges = edges
        self.el_flow_coeff_B: Optional[Dict[Tuple[str, str], float]] = None
        self.el_flow_coeff_DA: Optional[Dict[Tuple[str, str], float]] = None

    def define_constraints(self, pyomo_model: pyo.Model) -> None:
        super().define_constraints(pyomo_model=pyomo_model)

        if self.carrier_data.powerflow_method == "dc-pf":
            logger.warning("TODO: code for electric powerflow calculations need improvement (pu conversion)")
            nodelist = list(pyomo_model.setNode)  # self.all_nodes.keys()
            edgelist_el = {edge_id: asdict(edge.edge_data) for edge_id, edge in self.edges.items()}
            coeff_B, coeff_DA = el_calc.computePowerFlowMatrices(nodelist, edgelist_el, baseZ=1)
            self.el_flow_coeff_B = coeff_B
            self.el_flow_coeff_DA = coeff_DA

            # Reference voltage node:
            constr_ElVoltageReference = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_el_voltage_reference)
            setattr(
                pyomo_model,
                "constrN_{}_{}".format(self.carrier_data.reference_node, "voltageref"),
                constr_ElVoltageReference,
            )
            # Linearised power flow equations:
            for edge in self.edges.keys():
                constr_flow = pyo.Constraint(
                    edge,
                    pyomo_model.setHorizon,
                    rule=self._rule_dc_power_flow_equations,
                )
                setattr(pyomo_model, "constrE_{}_{}".format(edge, "flow"), constr_flow)

    def _rule_el_voltage_reference(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        el_carrier = self.carrier_data
        n = el_carrier.reference_node
        expr = pyomo_model.varElVoltageAngle[n, t] == 0
        return expr

    def _rule_dc_power_flow_equations(
        self, pyomo_model: pyo.Model, edge: dto.EdgeData, t: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Flow vs voltage angle"""
        base_mva = el_calc.elbase["baseMVA"]
        base_angle = el_calc.elbase["baseAngle"]
        lhs = pyomo_model.varEdgeFlow[edge, t]
        lhs = lhs / base_mva
        rhs = 0
        # TODO speed up creation of constraints - remove for loop
        n2s = [k[1] for k in self.el_flow_coeff_DA.keys() if k[0] == edge]
        for n2 in n2s:
            rhs += self.el_flow_coeff_DA[edge, n2] * (pyomo_model.varElVoltageAngle[n2, t] * base_angle)
        return lhs == rhs

    @staticmethod
    def compute_el_reserve(
        pyomo_model,
        t,
        all_devices: Dict[str, Device],
        exclude_device: str = None,
    ):
        """compute non-used generator capacity (and available loadflex)
        This is reserve to cope with forecast errors, e.g. because of wind
        variation or motor start-up

        all_devices : dict
            dictionary of all devices in the model {device_id:device_object}
        exclude_device : str (default None)
            compute reserve by devices excluding this one
        """
        alldevs = [d for d in pyomo_model.setDevice if d != exclude_device]
        # relevant devices are devices with el output or input
        cap_avail = 0
        p_generating = 0
        loadreduction = 0
        for d in alldevs:
            dev_obj = all_devices[d]
            reserve = dev_obj.compute_el_reserve(pyomo_model, t)
            cap_avail += reserve["capacity_available"]
            p_generating += reserve["capacity_used"]
            loadreduction += reserve["loadreduction_available"]

        res_dev = (cap_avail - p_generating) + loadreduction
        return res_dev



from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device
from oogeso.core.networks.network_node import NetworkNode


class Separator(Device):
    """Wellstream separation into oil/gas/water."""

    carrier_in = ["wellstream", "heat", "el"]
    carrier_out = ["oil", "gas", "water"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSeparatorData,
        carrier_data_dict: Dict[str, Union[dto.CarrierElData, dto.CarrierHeatData, dto.CarrierWellstreamData]],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rule_separator(self, pyomo_model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data = self.dev_data
        node = dev_data.node_id
        node_obj: NetworkNode = self.optimiser.all_nodes[node]  # Fixme: Unresolved reference
        wellstream_prop: dto.CarrierWellstreamData = self.all_carriers["wellstream"]  # Fixme: Unresolved reference
        GOR = wellstream_prop.gas_oil_ratio
        WC = wellstream_prop.water_cut
        comp_oil = (1 - WC) / (1 + GOR - GOR * WC)
        comp_water = WC / (1 + GOR - GOR * WC)
        comp_gas = GOR * (1 - WC) / (1 + GOR * (1 - WC))
        flow_in = pyomo_model.varDeviceFlow[dev, "wellstream", "in", t]
        if i == 1:
            lhs = pyomo_model.varDeviceFlow[dev, "gas", "out", t]
            rhs = flow_in * comp_gas
            return lhs == rhs
        elif i == 2:
            lhs = pyomo_model.varDeviceFlow[dev, "oil", "out", t]
            rhs = flow_in * comp_oil
            return lhs == rhs
        elif i == 3:
            # return pyo.Constraint.Skip
            lhs = pyomo_model.varDeviceFlow[dev, "water", "out", t]
            rhs = flow_in * comp_water
            return lhs == rhs
        elif i == 4:
            # electricity demand
            lhs = pyomo_model.varDeviceFlow[dev, "el", "in", t]
            rhs = flow_in * dev_data.el_demand_factor
            return lhs == rhs
        elif i == 5:
            lhs = pyomo_model.varDeviceFlow[dev, "heat", "in", t]
            rhs = flow_in * dev_data.heat_demand_factor
            return lhs == rhs
        elif i == 6:
            # gas pressure out = nominal
            lhs = pyomo_model.varPressure[(node, "gas", "out", t)]
            rhs = node_obj.get_pressure_nominal("gas", "out")
            return lhs == rhs
        elif i == 7:
            # oil pressure out = nominal
            lhs = pyomo_model.varPressure[(node, "oil", "out", t)]
            rhs = node_obj.get_pressure_nominal("oil", "out")
            return lhs == rhs
        elif i == 8:
            # water pressure out = nominal
            lhs = pyomo_model.varPressure[(node, "water", "out", t)]
            rhs = node_obj.get_pressure_nominal("water", "out")
            return lhs == rhs
        else:
            raise ValueError(f"Argument i must be 1, 2, ..., 8. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_separator = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 8), rule=self._rule_separator)
        # add constraints to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "separator"),
            constr_separator,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        raise NotImplementedError("No get_flow_var defined for Separator")


class Separator2(Device):
    "Wellstream separation into oil/gas/water"

    carrier_in = ["oil", "gas", "water", "heat", "el"]
    carrier_out = ["oil", "gas", "water"]
    serial = ["oil", "gas", "water"]

    def __init__(
        self,
        dev_data: dto.DeviceSeparator2Data,
        carrier_data_dict: Dict[
            str,
            Union[dto.CarrierGasData, dto.CarrierOilData, dto.CarrierElData, dto.CarrierHeatData, dto.CarrierWaterData],
        ],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    # Alternative separator model - using oil/gas/water input instead of
    # wellstream
    def _rule_separator2_flow(
        self, pyomo_model: pyo.Model, fc, t: int, i: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        node = self.dev_data.node_id
        node_obj: NetworkNode = self.node
        # wellstream_prop=self.pyomo_model.all_carriers['wellstream']
        # flow_in = sum(model.varDeviceFlow[dev,f,'in',t]
        #                for f in['oil','gas','water'])
        if i == 1:
            # component flow in = flow out
            lhs = pyomo_model.varDeviceFlow[dev, fc, "out", t]
            rhs = pyomo_model.varDeviceFlow[dev, fc, "in", t]
            return lhs == rhs
        elif i == 2:
            # pressure out is nominal
            lhs = pyomo_model.varPressure[(node, fc, "out", t)]
            rhs = node_obj.get_pressure_nominal(fc, "out")
            return lhs == rhs
        else:
            raise ValueError(f"Argument i must be 1 or 2. {i} was given.")

    def _rule_separator2_energy(
        self, pyomo_model: pyo.Model, t: int, i: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data: dto.DeviceSeparator2Data = self.dev_data
        flow_in = sum(pyomo_model.varDeviceFlow[dev, f, "in", t] for f in ["oil", "gas", "water"])

        if i == 1:
            # electricity demand
            lhs = pyomo_model.varDeviceFlow[dev, "el", "in", t]
            rhs = flow_in * dev_data.el_demand_factor
            return lhs == rhs
        elif i == 2:
            # heat demand
            lhs = pyomo_model.varDeviceFlow[dev, "heat", "in", t]
            rhs = flow_in * dev_data.heat_demand_factor
            return lhs == rhs
        else:
            raise ValueError(f"Argument i must be 1 or 2. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""
        # No specific constraints, use only generic ones:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_separator2_flow = pyo.Constraint(
            ["oil", "gas", "water"],
            pyomo_model.setHorizon,
            pyo.RangeSet(1, 2),
            rule=self._rule_separator2_flow,
        )
        # add constraints to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "flow"),
            constr_separator2_flow,
        )

        constr_separator2_energy = pyo.Constraint(
            pyomo_model.setHorizon,
            pyo.RangeSet(1, 2),
            rule=self._rule_separator2_energy,
        )
        # add constraints to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "energy"),
            constr_separator2_energy,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        raise NotImplementedError("No get_flow_var defined for Separator")



import logging

import numpy as np
import pandas as pd
import pydot
import pyomo.environ as pyo

try:
    import matplotlib.pyplot as plt
    import plotly
    import plotly.express as px
    import seaborn as sns
except ImportError:
    raise ImportError("In order to run this plotting module you need to install matplotlib, plotly and seaborn.")

sns.set_style("whitegrid")  # Optional: sns.set_palette("dark")

plotter = "plotly"  # matplotlib

logger = logging.getLogger(__name__)


def plot_df(df, id_var, filename=None, title=None, ylabel="value"):
    """Plot dataframe using plotly (saved to file)"""

    df_tidy = df.reset_index()
    df_tidy.rename(columns={0: ylabel}, inplace=True)
    fig = px.line(df_tidy, x="time", y=ylabel, color=id_var, title=title)

    # fig = px.line()
    # for cols in df:
    #    fig.add_scatter(df[cols])
    # fig = px.line(df,x=df.index,y=df.values,title=title)
    # py.iplot('data':[{'x':df.index,'y':df.values,'type':'line'}],
    #        'layout': {'title':'title'},
    #        filename=filename)
    if filename is not None:
        plotly.offline.plot(fig, filename=filename)

    return fig


def plot_device_profile(
    sim_result,
    optimisation_model,
    devs,
    filename=None,
    reverse_legend=True,
    include_forecasts=False,
    include_on_off=False,
    include_prep=False,
    devs_shareload=None,
):
    """plot forecast and actual profile (available power), and device output

    Parameters
    ==========
    sim_results : SimulationResults object
    optimisation_model : OptimisationModel object
    devs : list
        which devices to include
    devs_shareload : list ([]=ignore, None=do it for gas turbines)
        list of devices for which displayed power should be shared evenly
        (typically gas turbines)
        The optimision returns somewhat random distribution of load per device,
        in reality they will share load more or less evenly due to their
        frequency droop settings. Rather than imposing this in the optimisation,
        this is included in the plots. Default: gas turbine

    """
    res = sim_result
    optimiser = optimisation_model
    if type(devs) is not list:
        devs = [devs]
    if include_forecasts & (len(devs) > 1):
        print("Can only plot one device when showing forecasts")
        return
    df = res.device_flow.unstack(["carrier", "terminal"])[("el", "out")].unstack("device")
    if devs_shareload is None:
        # gas turbines:
        devs_shareload = [d for d, d_obj in optimiser.all_devices.items() if d_obj.dev_data.model == "gasturbine"]
        # devs_shareload = [d for d in mc.instance.setDevice
        #    if mc.instance.paramDevice[d]['model']=='gasturbine']
    if devs_shareload:  # list is non-empty
        devs_online = (df[devs_shareload] > 0).sum(axis=1)
        devs_sum = df[devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in devs_shareload:
            mask = df[c] > 0
            df.loc[mask, c] = devs_mean[mask]
    df.columns.name = "devices"
    df = df[devs]
    nrows = 1
    if include_on_off:
        nrows = nrows + 1
    if include_prep:
        nrows = nrows + 1
    df2 = res.device_is_on.unstack("device")[devs]
    dfPrep = res.device_is_prep.unstack("device")[devs]
    timerange = list(res.device_is_on.index.get_level_values("time"))
    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=nrows, cols=1, shared_xaxes=True)
        colour = plotly.colors.DEFAULT_PLOTLY_COLORS
        k = 0
        rowOnOff = 2
        rowPrep = 2
        if include_on_off:
            rowPrep = 3
        for col in df:
            dev = col
            dev_data = optimiser.all_devices[dev].dev_data
            k = (k + 1) % len(colour)  # repeat colour cycle if necessary
            fig.add_scatter(
                x=df.index,
                y=df[col],
                line_shape="hv",
                name=col,
                line=dict(color=colour[k]),
                stackgroup="P",
                legendgroup=col,
                row=1,
                col=1,
            )
            if include_on_off & (dev_data.start_stop is not None):
                fig.add_scatter(
                    x=df2.index,
                    y=df2[col],
                    line_shape="hv",
                    name=col,
                    line=dict(color=colour[k], dash="dash"),
                    stackgroup="ison",
                    legendgroup=col,
                    row=rowOnOff,
                    col=1,
                    showlegend=False,
                )
            if include_prep & (dev_data.start_stop is not None):
                fig.add_scatter(
                    x=dfPrep.index,
                    y=dfPrep[col],
                    line_shape="hv",
                    name=col,
                    line=dict(color=colour[k], dash="dash"),
                    stackgroup="ison",
                    legendgroup=col,
                    row=rowPrep,
                    col=1,
                    showlegend=False,
                )
            if include_forecasts & (dev_data.profile is not None):
                curve = dev_data.profile
                devPmax = dev_data.flow_max
                if curve in res.profiles_nowcast:
                    fig.add_scatter(
                        x=timerange,
                        y=res.profiles_nowcast.loc[timerange, curve] * devPmax,
                        line_shape="hv",
                        line=dict(color=colour[k + 1]),
                        name="--nowcast",
                        legendgroup=col,
                        row=1,
                        col=1,
                    )
                fig.add_scatter(
                    x=timerange,
                    y=res.profiles_forecast.loc[timerange, curve] * devPmax,
                    line_shape="hv",
                    line=dict(color=colour[k + 2]),
                    name="--forecast",
                    legendgroup=col,
                    row=1,
                    col=1,
                )
        fig.update_xaxes(row=1, col=1, title_text="")
        fig.update_xaxes(row=nrows, col=1, title_text="Timestep")
        fig.update_yaxes(row=1, col=1, title_text="Power supply (MW)")
        if include_on_off:
            fig.update_yaxes(row=rowOnOff, col=1, title_text="On/off status")
        if include_prep:
            fig.update_yaxes(row=rowPrep, col=1, title_text="Startup", nticks=2)
        if reverse_legend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
        # fig.show()
    elif plotter == "matplotlib":
        fig, axs = plt.subplots(nrows=nrows, ncols=1, shared_xaxes=True, figsize=(12, 1 + 3 * nrows))
        #
        #        fig = plt.figure()
        ax = axs[0]
        labels = []
        offset_online = 0
        # df.plot(ax=ax)
        for dev in devs:
            # dev_param = mc.instance.paramDevice[dev]
            dev_data = optimiser.all_devices[dev].dev_data
            devname = "{}:{}".format(dev, dev_data.name)
            devPmax = dev_data.flow_max
            df[dev].plot(ax=ax)
            # get the color of the last plotted line (the one just plotted)
            col = ax.get_lines()[-1].get_color()
            labels = labels + [devname]
            if include_forecasts & (dev_data.profile is not None):
                curve = dev_data.profile
                (res.profiles_nowcast.loc[timerange, curve] * devPmax).plot(ax=ax, linestyle="--")
                # ax.set_prop_cycle(None)
                (res.profiles_forecast.loc[timerange, curve] * devPmax).plot(ax=ax, linestyle=":")
                labels = labels + ["--nowcast", "--forecast"]
            if include_on_off & (dev_data.start_stop is not None):
                # df2=res.dfDeviceIsOn.unstack(0)[dev]+offset_online
                offset_online += 0.1
                df2[dev].plot(ax=ax, linestyle="--", color=col)
                labels = labels + ["--online"]
        plt.xlim(df.index.min(), df.index.max())
        ax.legend(labels, loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_device_power_energy(sim_result, optimisation_model: pyo.Model, dev, filename=None, energy_fill_opacity=None):
    """Plot power in/out of device and storage level (if any)"""
    res = sim_result
    optimiser = optimisation_model
    dev_data = optimiser.all_devices[dev].dev_data
    devname = "{}:{}".format(dev, dev_data.name)

    if dev_data.model == "storage_hydrogen":
        carrier = "hydrogen"
        Ptitle = "Flow (Sm3/s)"
        Etitle = "Energy storage( Sm3)"
    else:
        carrier = "el"
        Ptitle = "Power (MW)"
        Etitle = "Energy storage (MWh)"
    # Power flow in/out
    dfF = res.device_flow[dev, carrier].unstack("terminal")
    dfE = res.device_storage_energy.unstack("device")
    # dfE = pd.DataFrame()
    if dev in dfE:
        dfE = dfE[dev]
        # Shift time by one, since the dfDeviceEnergy[t] is the energy _after_
        # timestep t:
        #        dfE = dfE.unstack().T
        dfE.index = dfE.index + 1
        # dfE = dfE.rename(columns={0: "storage"})
        # dfE.loc[dfE.index.min()-1,dev] = mc.instance.param

    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(specs=[[{"secondary_y": True}]])
        for col in dfF.columns:
            fig.add_scatter(
                x=dfF.index,
                y=dfF[col],
                line_shape="hv",
                name=col,
                secondary_y=True,
                fill="tozeroy",
            )
        if not dfE.empty:
            fig.add_scatter(
                x=dfE.index,
                y=dfE,
                name="storage",
                secondary_y=False,
                fill="tozeroy",
            )  # ,line=dict(dash='dot'))
            if energy_fill_opacity is not None:
                k = len(fig["data"]) - 1
                linecol = plotly.colors.DEFAULT_PLOTLY_COLORS[k]
                opacity = energy_fill_opacity
                fillcol = "rgba({}, {})".format(linecol[4:][:-1], opacity)
                fig["data"][k]["fillcolor"] = fillcol
                fig["data"][k]["fill"] = "tozeroy"
            fig.update_yaxes(title_text=Etitle, secondary_y=False, side="right")
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text=Ptitle, secondary_y=True, side="left")

    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        plt.title(devname)
        ax = plt.gca()
        dfF.plot(ax=ax, drawstyle="steps-post", marker=".")
        ax.set_xlabel("Timestep")
        ax.set_ylabel(Ptitle)
        tmin = dfF.index.get_level_values("time").min()
        tmax = dfF.index.get_level_values("time").max() + 1
        ax.set_ylim(0, dev_data.flow_max)
        ax.legend(loc="upper left")  # , bbox_to_anchor =(1.01,0),frameon=False)

        if not dfE.empty:
            ax2 = ax.twinx()
            ax2.grid(None)
            dfE.plot(ax=ax2, linestyle=":", color="black")
            ax2.set_ylabel("Energy (MWh)")  # ,color="red")
            if dev_data.model in ["storage_el"]:
                ax2.set_ylim(0, dev_data.max_E)
            elif dev_data.model in ["well_injection"]:
                ax2.set_ylim(-dev_data.max_E / 2, dev_data.max_E / 2)
            # ax2.tick_params(axis='y', labelcolor="red")
            ax2.legend(loc="upper right")
        ax.set_xlim(tmin, tmax)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_sum_power_mix(
    sim_result,
    optimisation_model,
    carrier,
    filename=None,
    reverse_legend=True,
    exclude_zero=False,
    devs_shareload=None,
):
    """
    Plot power mix

    Parameters
    ==========
    sim_result : SimulationResult object
    optimisation_model : Oogeso optimisation model
    carrier : string
    devs_shareload : list ([]=ignore, None=do it for gas turbines)
        list of devices for which power should be shared evenly (typically gas turbines)
        The optimision returns somewhat random distribution of load per device,
        in reality they will share load more or less evenly due to their
        frequency droop settings. Rather than imposing this in the optimisation,
        this is included in the plots.
    """
    # optimiser = simulator.optimiser
    res = sim_result
    optimiser = optimisation_model
    # Power flow in/out
    dfF = res.device_flow
    tmin = dfF.index.get_level_values("time").min()
    tmax = dfF.index.get_level_values("time").max() + 1
    mask_carrier = dfF.index.get_level_values("carrier") == carrier
    mask_in = dfF.index.get_level_values("terminal") == "in"
    mask_out = dfF.index.get_level_values("terminal") == "out"
    dfF_out = dfF[mask_carrier & mask_out]
    dfF_out.index = dfF_out.index.droplevel(level=("carrier", "terminal"))
    dfF_out = dfF_out.unstack(0)
    keepcols = optimiser.getDevicesInout(carrier_out=carrier)
    # logger.info("in: {}".format(keepcols))
    dfF_out = dfF_out[keepcols]
    dfF_in = dfF[mask_carrier & mask_in]
    dfF_in.index = dfF_in.index.droplevel(level=("carrier", "terminal"))
    dfF_in = dfF_in.unstack(0)
    keepcols = optimiser.getDevicesInout(carrier_in=carrier)
    # logger.info("out: {}".format(keepcols))
    dfF_in = dfF_in[keepcols]

    if (devs_shareload is None) and (carrier in ["el", "heat"]):
        # gas turbines:
        devs_shareload = [d for d, d_obj in optimiser.all_devices.items() if d_obj.dev_data.model == "gasturbine"]
        logger.debug("Shared load=%s", devs_shareload)
    if devs_shareload:  # list is non-empty
        devs_online = (dfF_out[devs_shareload] > 0).sum(axis=1)
        devs_sum = dfF_out[devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in devs_shareload:
            mask = dfF_out[c] > 0
            dfF_out.loc[mask, c] = devs_mean[mask]

    if exclude_zero:
        dfF_in = dfF_in.loc[:, dfF_in.sum() != 0]
        dfF_out = dfF_out.loc[:, dfF_out.sum() != 0]

    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.05)
        for col in dfF_in:
            fig.add_scatter(
                x=dfF_in.index,
                y=dfF_in[col],
                line_shape="hv",
                name="in:" + col,
                stackgroup="in",
                legendgroup=col,
                row=2,
                col=1,
            )
        for col in dfF_out:
            fig.add_scatter(
                x=dfF_out.index,
                y=dfF_out[col],
                line_shape="hv",
                name="out:" + col,
                stackgroup="out",
                legendgroup=col,
                row=1,
                col=1,
            )
        fig.update_xaxes(row=1, col=1, title_text="")
        fig.update_xaxes(row=2, col=1, title_text="Timestep")
        fig.update_yaxes(row=1, col=1, title_text="Power supply (MW)")
        fig.update_yaxes(row=2, col=1, title_text="Power consumption (MW)")
        if reverse_legend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
        # fig.show()
    elif plotter == "matplotlib":
        fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(12, 8))
        # plt.figure(figsize=(12,4))
        plt.suptitle("Sum power ({})".format(carrier))
        dfF_out.plot.area(ax=axes[0], linewidth=0)
        dfF_in.plot.area(ax=axes[1], linewidth=0)
        axes[0].set_ylabel("Power supply (MW)")
        axes[1].set_ylabel("Power consumption (MW)")
        axes[0].set_xlabel("")
        axes[1].set_xlabel("Timestep")
        for ax in axes:
            if reverse_legend:
                handles, labels = ax.get_legend_handles_labels()
                ax.legend(
                    handles[::-1],
                    labels[::-1],
                    loc="lower left",
                    bbox_to_anchor=(1.01, 0),
                    frameon=False,
                )
            else:
                ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
            ax.set_xlim(tmin, tmax)

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_export_revenue(sim_result, filename=None, currency="$"):
    export_revenue = sim_result.export_revenue.unstack("carrier")
    if plotter == "plotly":
        dfplot = export_revenue.loc[:, export_revenue.sum() > 0]
        fig = px.area(dfplot)
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Revenue ({}/s)".format(currency))
    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        plt.title("Export revenue ($/s)")
        ax = plt.gca()
        ax.set_ylabel("{}/s".format(currency))
        ax.set_xlabel("Timestep")
        (export_revenue.loc[:, export_revenue.sum() > 0]).plot.area(ax=ax, linewidth=0)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_CO2_rate(sim_result, filename=None):
    plt.figure(figsize=(12, 4))
    plt.title("CO2 emission rate (kgCO2/s)")
    ax = plt.gca()
    ax.set_ylabel("kgCO2/s")
    ax.set_xlabel("Timestep")
    sim_result.co2_rate.plot()
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_CO2_rate_per_dev(
    sim_result,
    optimisation_model,
    filename=None,
    reverseLegend=False,
    devs_shareload=[],
):

    # df_info = pd.DataFrame.from_dict(dict(mc.instance.paramDevice.items())).T
    # labels = (df_info.index.astype(str))# +'_'+df_info['name'])
    dfco2rate = sim_result.co2_rate_per_dev.unstack("device")
    all_devices = optimisation_model.all_devices
    dfplot = dfco2rate.loc[:, ~(dfco2rate == 0).all()].copy()

    if devs_shareload is None:
        # gas turbines:
        devs_shareload = [d for d, d_obj in all_devices.items() if d_obj.dev_data.model == "gasturbine"]
    #        devs_shareload = [d for d in mc.instance.setDevice
    #            if ((mc.instance.paramDevice[d]['model']=='gasturbine') and
    #                (d in dfplot))]
    if devs_shareload:  # list is non-empty
        devs_shareload = [d for d in devs_shareload if d in dfplot]
        devs_online = (dfplot[devs_shareload] > 0).sum(axis=1)
        devs_sum = dfplot[devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in devs_shareload:
            mask = dfplot[c] > 0
            dfplot.loc[mask, c] = devs_mean[mask]

    # dfplot.columns=labels
    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=1, cols=1)
        # ,shared_xaxes=True,vertical_spacing=0.05)
        # fig = px.line(dfplot,x="time",y=ylabel,color=id_var,title=title)
        for col in dfplot:
            fig.add_scatter(
                x=dfplot.index,
                y=dfplot[col],
                line_shape="hv",
                name=col,
                stackgroup="one",
                row=1,
                col=1,
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Emission rate (kgCO2/s)")
        if reverseLegend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
    else:
        fig = plt.figure(figsize=(12, 4))
        ax = plt.gca()
        ax.set_ylabel("Emission rate (kgCO2/s)")
        ax.set_xlabel("Timestep")
        dfplot.plot.area(ax=ax, linewidth=0)
        if reverseLegend:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(
                handles[::-1],
                labels[::-1],
                loc="lower left",
                bbox_to_anchor=(1.01, 0),
                frameon=False,
            )
        else:
            ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_CO2_intensity(sim_result, filename=None):
    title = "CO2 intensity (kgCO2/Sm3oe)"
    xlabel = "Timestep"
    ylabel = "CO2 intensity (kgCO2/Sm3oe)"
    dfplot = sim_result.co2_intensity
    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=1, cols=1)
        # ,shared_xaxes=True,vertical_spacing=0.05)
        fig = px.line(dfplot, x=dfplot.index, y=dfplot.values)  # ,title=title)
        fig.update_xaxes(title_text=xlabel)
        fig.update_yaxes(title_text=ylabel)
    else:
        fig = plt.figure(figsize=(12, 4))
        plt.title(title)
        ax = plt.gca()
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        dfplot.plot()
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_profiles(profiles, filename=None):
    """Plot profiles (forecast and actual)"""
    fig = None
    if isinstance(profiles, list):
        # list of TimeSeriesData objects
        df = pd.DataFrame()
        for d in profiles:
            df[(d.id, "forecast")] = d.data
            if d.data_nowcast is not None:
                df[(d.id, "nowcast")] = d.data_nowcast
        df.index.name = "timestep"
        df.columns = pd.MultiIndex.from_tuples(df.columns, names=("variable", "type"))
        # df = df.stack("type")
    elif isinstance(profiles, dict):
        df = pd.concat(
            {
                "nowcast": profiles["actual"],
                "forecast": profiles["forecast"],
            }
        )
        df.index.names = ["type", "timestep"]
    else:
        raise Exception("profiles must be input data format or internal format")
    if plotter == "plotly":
        df = df.reset_index()
        df = df.melt(id_vars=["timestep"])
        fig = px.line(
            df,
            x="timestep",
            y="value",
            line_group="type",
            color="variable",
            line_dash="type",
            line_shape="hv",
        )
    elif plotter == "matplotlib":
        plt.figure(figsize=(12, 4))
        ax = plt.gca()
        df.loc[:, df.columns.get_level_values("type") == "forecast"].plot(ax=ax)
        # reset color cycle (so using the same as for the actual plot):
        ax.set_prop_cycle(None)
        df.loc[:, df.columns.get_level_values("type") == "nowcast"].plot(ax=ax, linestyle=":")
        ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        plt.xlabel("Timestep")
        plt.ylabel("Relative value")
        plt.title("Forecast and nowcast profile (forecast=sold line)")
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_device_power_flow_pressure(sim_result, optimisation_model: pyo.Model, dev, carriers_inout=None, filename=None):
    res = sim_result
    all_devices = optimisation_model.all_devices
    dev_obj = all_devices[dev]
    dev_data = dev_obj.dev_data
    node = dev_data.node_id
    devname = "{}:{}".format(dev, dev_data.name)
    # linecycler = itertools.cycle(['-','--',':','-.']*10)
    if carriers_inout is None:
        carriers_inout = {"in": dev_obj.carrier_in, "out": dev_obj.carrier_out}
    if "serial" in carriers_inout:
        del carriers_inout["serial"]

    plt.figure(figsize=(12, 4))
    ax = plt.gca()
    # ax.plot(res.dfDevicePower.unstack(0)[dev],'-.',label="DevicePower")
    for inout, carriers in carriers_inout.items():
        if inout == "in":
            ls = "--"
        else:
            ls = ":"
        for carr in carriers:
            ax.plot(
                res.device_flow.unstack([0, 1, 2])[(dev, carr, inout)],
                ls,
                label="DeviceFlow ({},{})".format(carr, inout),
            )
    # Pressure
    for inout, carriers in carriers_inout.items():
        for carr in carriers:
            if carr != "el":
                ax.plot(
                    res.terminal_pressure.unstack(0)[node].unstack([0, 1])[(carr, inout)],
                    label="TerminalPressure ({},{})".format(carr, inout),
                )
    plt.title(devname)
    plt.xlabel("Timestep")
    plt.ylabel("Value")
    plt.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_network(
    simulator,
    timestep=None,
    filename=None,
    prog="dot",
    only_carrier=None,
    rankdir="LR",
    plotDevName=False,
    numberformat="{:.2g}",
    hide_losses=False,
    hide_edgelabel=False,
    **kwargs
):
    """Plot energy network

    simulator : Simulation object
    timestep : int
        which timestep to show values for
    filename : string
        Name of file
    only_carrier : str or list
        Restrict energy carriers to these types (None=plot all)
    rankdir : str
        Plotting direction TB=top to bottom, LR=left to right
    numberformat : str
        specify how numbers should be represented in plot
    hide_losses : bool
        Don't show losses on edges (if any)
    hide_edgelabel : bool
        Don't show any labels on edges
    **kwargs :
        Additional arguments passed on to pydot.Dot(...)
    """

    # Idea: in general, there are "in" and "out" terminals. If there are
    # no serial devices, then these are merged into a single terminal
    # (prettier plot"). Whether the single terminal is shown as an in or out
    # terminal (left or irght), depends on whether it is an input or output
    # of a majority of the connected devices.

    optimiser = simulator.optimiser
    model = optimiser
    res = simulator.result_object

    cluster = {}
    col = {
        "t": {
            "el": "red",
            "gas": "orange",
            "heat": "darkgreen",
            "wellstream": "brown",
            "oil": "black",
            "water": "blue4",
            "hydrogen": "deepskyblue2",
        },
        "e": {
            "el": "red",
            "gas": "orange",
            "heat": "darkgreen",
            "wellstream": "brown",
            "oil": "black",
            "water": "blue4",
            "hydrogen": "deepskyblue2",
        },
        "d": "white",
        "cluster": "lightgray",
    }
    # dotG = pydot.Dot(graph_type='digraph') #rankdir='LR',newrank='false')
    dotG = pydot.Dot(graph_type="digraph", rankdir=rankdir, **kwargs)
    if only_carrier is None:
        carriers = model.setCarrier
    elif type(only_carrier) is str:
        carriers = [only_carrier]
    else:
        carriers = only_carrier

    # devicemodels = milp_compute.devicemodel_inout()

    # plot all node and terminals:
    for n_id, node_obj in optimiser.all_nodes.items():
        cluster = pydot.Cluster(graph_name=n_id, label=n_id, style="filled", color=col["cluster"])
        terms_in = pydot.Subgraph(rank="min")
        gr_devices = pydot.Subgraph(rank="same")
        terms_out = pydot.Subgraph(rank="max")
        for carrier in carriers:
            # add only terminals that are connected to something (device or edge)
            if node_obj.is_non_trivial(carrier):
                devs = node_obj.devices
                num_in = 0
                num_out = 0
                for d, dev_obj in devs.items():
                    dev_model = dev_obj.dev_data.model
                    devlabel = d  # use index as label
                    devlabel = "{}\n{}".format(d, dev_model)
                    if plotDevName:
                        dev_name = dev_obj.dev_data.name
                        devlabel = "{} {}".format(devlabel, dev_name)
                    carriers_in = dev_obj.carrier_in
                    carriers_out = dev_obj.carrier_out
                    carriers_in_lim = list(set(carriers_in) & set(carriers))
                    carriers_out_lim = list(set(carriers_out) & set(carriers))
                    if (carriers_in_lim != []) or (carriers_out_lim != []):
                        gr_devices.add_node(pydot.Node(d, color=col["d"], style="filled", label=devlabel))
                    if carrier in carriers_in_lim:
                        num_in += 1
                        if timestep is None:
                            devedgelabel = ""
                        else:
                            f_in = res.device_flow[(d, carrier, "in", timestep)]
                            devedgelabel = numberformat.format(f_in)
                        if carrier in node_obj.devices_serial:
                            n_in = n_id + "_" + carrier + "_in"
                        else:
                            n_in = n_id + "_" + carrier
                        dotG.add_edge(
                            pydot.Edge(
                                dst=d,
                                src=n_in,
                                color=col["e"][carrier],
                                fontcolor=col["e"][carrier],
                                label=devedgelabel,
                            )
                        )
                    if carrier in carriers_out_lim:
                        num_out += 1
                        if timestep is None:
                            devedgelabel = ""
                        else:
                            f_out = res.device_flow[(d, carrier, "out", timestep)]
                            devedgelabel = numberformat.format(f_out)
                        if carrier in node_obj.devices_serial:
                            n_out = n_id + "_" + carrier + "_out"
                        else:
                            n_out = n_id + "_" + carrier
                        dotG.add_edge(
                            pydot.Edge(
                                dst=n_out,
                                src=d,
                                color=col["e"][carrier],
                                fontcolor=col["e"][carrier],
                                label=devedgelabel,
                            )
                        )

                # add in/out terminals
                supp = ""
                if carrier in node_obj.devices_serial:
                    supp = "_out"  # noqa
                label_in = carrier  # + "_in "
                label_out = carrier  # + supp + " "
                if timestep is None:
                    pass
                elif carrier in ["gas", "wellstream", "oil", "water"]:
                    label_in += numberformat.format(res.terminal_pressure[(n_id, carrier, "in", timestep)])
                    label_out += numberformat.format(res.terminal_pressure[(n_id, carrier, "out", timestep)])
                elif carrier == "el":
                    if optimiser.all_networks["el"].carrier_data.powerflow_method == "dc-pf":
                        label_in += numberformat.format(res.el_voltage_angle[(n_id, timestep)])
                        label_out += numberformat.format(res.el_voltage_angle[(n_id, timestep)])
                # Add two terminals if there are serial devices, otherwise one:
                if carrier in node_obj.devices_serial:
                    terms_in.add_node(
                        pydot.Node(
                            name=n_id + "_" + carrier + "_in",
                            color=col["t"][carrier],
                            label=label_in,
                            shape="box",
                        )
                    )
                    terms_out.add_node(
                        pydot.Node(
                            name=n_id + "_" + carrier + "_out",
                            color=col["t"][carrier],
                            label=label_out,
                            shape="box",
                        )
                    )
                else:
                    # TODO: make this in or out depending on connected devices
                    if num_out > num_in:
                        terms_out.add_node(
                            pydot.Node(
                                name=n_id + "_" + carrier,
                                color=col["t"][carrier],
                                label=label_out,
                                shape="box",
                            )
                        )
                    else:
                        terms_in.add_node(
                            pydot.Node(
                                name=n_id + "_" + carrier,
                                color=col["t"][carrier],
                                label=label_out,
                                shape="box",
                            )
                        )

        cluster.add_subgraph(terms_in)
        cluster.add_subgraph(gr_devices)
        cluster.add_subgraph(terms_out)
        dotG.add_subgraph(cluster)

    # plot all edges (per carrier):
    for carrier in carriers:
        for i, edge_obj in optimiser.all_edges.items():
            edge_data = edge_obj.edge_data
            if edge_data.carrier == carrier:
                headlabel = ""
                taillabel = ""
                if hide_edgelabel:
                    edgelabel = ""
                elif timestep is None:
                    edgelabel = ""
                    if (not hide_edgelabel) and hasattr(edge_data, "pressure_from"):
                        edgelabel = "{} {}-{}".format(
                            edgelabel,
                            edge_data.pressure_from,
                            edge_data.pressure_to,
                        )
                        # edgelabel = "{}-{}".format(edgelabel, edge_data.pressure["to"])
                else:
                    edgelabel = numberformat.format(res.dfEdgeFlow[(i, timestep)])
                    # Add loss
                    if (not hide_losses) and (res.edge_loss is not None) and ((i, timestep) in res.edge_loss):
                        # taillabel = " " + edgelabel
                        losslabel = numberformat.format(res.edge_loss[(i, timestep)])
                        edgelabel = "{} [{}]".format(edgelabel, losslabel)
                n_from = edge_data.node_from
                n_to = edge_data.node_to
                n_from_obj = optimiser.all_nodes[n_from]
                n_to_obj = optimiser.all_nodes[n_to]
                # name of terminal depends on whether it serial or single
                if carrier in n_from_obj.devices_serial:
                    t_out = n_from + "_" + carrier + "_out"
                else:
                    t_out = n_from + "_" + carrier
                if carrier in n_to_obj.devices_serial:
                    t_in = n_to + "_" + carrier + "_in"
                else:
                    t_in = n_to + "_" + carrier

                dotG.add_edge(
                    pydot.Edge(
                        src=t_out,
                        dst=t_in,
                        color='"{0}:invis:{0}"'.format(col["e"][carrier]),
                        fontcolor=col["e"][carrier],
                        label=edgelabel,
                        taillabel=taillabel,
                        headlabel=headlabel,
                    )
                )

    if filename is not None:
        # prog='dot' gives the best layout.
        dotG.write_png(filename, prog=prog)
    return dotG


def plot_gas_turbine_efficiency(fuelA=2.35, fuelB=0.53, energycontent=40, co2content=2.34, filename=None, Pmax=None):
    """
    co2content : CO2 content, kgCO2/Sm3gas
    energycontent: energy content, MJ/Sm3gas
    A,B : linear parameters
    """

    x_pow = np.linspace(0, 1, 50)
    y_fuel = fuelB + fuelA * x_pow
    # Pgas = Qgas*energycontent: Qgas=Pgas/Pmax * Pmax/energycontent

    nplots = 3 if Pmax is None else 4
    # if Pmax is not None
    plt.figure(figsize=(4 * nplots, 4))
    # plt.suptitle("Gas turbine fuel characteristics")

    if Pmax is not None:
        y_fuel_sm3 = y_fuel * Pmax / energycontent  # Sm3/s
        plt.subplot(1, nplots, 1)
        plt.title("Fuel usage (Sm3/h)")
        plt.xlabel("Electric power output (MW)")
        plt.plot(x_pow * Pmax, y_fuel_sm3 * 3600)  # per hour
        plt.ylim(bottom=0)

    plt.subplot(1, nplots, nplots - 2)
    plt.title("Fuel usage ($P_{gas}/P_{el}^{max}$)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    # plt.ylabel("Gas power input ($P_{gas}/P_{el}^{max}$)")
    plt.plot(x_pow, y_fuel)
    plt.ylim(bottom=0)

    plt.subplot(1, nplots, nplots - 1)
    plt.title("Efficiency ($P_{el}/P_{gas}$)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    plt.plot(x_pow, x_pow / y_fuel)

    #    plt.subplot(1,3,3)
    #    plt.title("Specific fuel usage ($P_{gas}/P_{el}$)")
    #    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    #    plt.plot(x_pow,y_fuel/x_pow)
    #    plt.ylim(top=30)

    # 1 MJ = 3600 MWh
    with np.errstate(divide="ignore", invalid="ignore"):
        emissions = 3600 * co2content / energycontent * y_fuel / x_pow
    plt.subplot(1, nplots, nplots)
    plt.title("Emission intensity (kgCO2/MWh)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    plt.plot(x_pow, emissions)
    plt.ylim(top=2000)

    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_reserve(
    sim_result,
    optimisation_model,
    includeMargin=True,
    dynamicMargin=True,
    useForecast=False,
    includeSum=True,
):
    """Plot unused online capacity by all el devices"""
    df_devs = pd.DataFrame()
    res = sim_result
    optimiser = optimisation_model
    timerange = list(res.el_reserve.index)
    marginIncr = pd.DataFrame(0, index=timerange, columns=["margin"])
    for d, dev_obj in optimiser.all_devices.items():
        dev_data = dev_obj.dev_data
        devmodel = dev_data.model
        rf = 1
        if "el" in dev_obj.carrier_out:
            # Generators and storage
            maxValue = dev_data.flow_max
            if dev_data.profile is not None:
                extprofile = dev_data.profile
                if useForecast or (extprofile not in res.profiles_nowcast):
                    maxValue = maxValue * res.profiles_forecast.loc[timerange, extprofile]
                else:
                    maxValue = maxValue * res.profiles_nowcast.loc[timerange, extprofile]
            if dev_data.start_stop is not None:  # devmodel in ["gasturbine"]:
                ison = res.device_is_on[d]
                maxValue = ison * maxValue
            elif devmodel in ["storage_el"]:
                maxValue = res.dfDeviceStoragePmax[d] + res.device_flow[d, "el", "in"]
            if dev_data.reserve_factor is not None:
                reserve_factor = dev_data.reserve_factor
                if reserve_factor == 0:
                    # device does not count towards reserve
                    rf = 0
                if dynamicMargin:
                    # instead of reducing reserve, increase the margin instead
                    # R*0.8-M = R - (M+0.2R) - this shows better in the plot what
                    marginIncr["margin"] += rf * maxValue * (1 - reserve_factor)
                else:
                    maxValue = maxValue * reserve_factor
            cap_avail = rf * maxValue
            p_generating = rf * res.device_flow[d, "el", "out"]
            reserv = cap_avail - p_generating
            df_devs[d] = reserv
    df_devs.columns.name = "device"
    if plotter == "plotly":
        fig = px.area(df_devs, line_shape="hv")
        if includeSum:
            fig.add_scatter(
                x=df_devs.index,
                y=df_devs.sum(axis=1),
                name="SUM",
                line=dict(dash="dot", color="black"),
                line_shape="hv",
            )
        if includeMargin:
            margin = optimiser.all_networks["el"].carrier_data.el_reserve_margin
            # wind contribution (cf compute reserve)
            marginIncr["margin"] = marginIncr["margin"] + margin
            fig.add_scatter(
                x=marginIncr.index,
                y=marginIncr["margin"],
                name="Margin",
                line=dict(dash="dot", color="red"),
                mode="lines",
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Reserve (MW)")
        #    return df_devs,marginIncr
    elif plotter == "matplotlib":
        ax = df_devs.plot.area()
        if includeSum:
            df_devs.sum(axis=1).plot(style=":", color="black", drawstyle="steps-post")
        if includeMargin:
            margin = optimiser.all_networks["el"].carrier_data.el_reserve_margin
            # wind contribution (cf compute reserve)
            marginIncr["margin"] = marginIncr["margin"] + margin
            marginIncr[["margin"]].plot(style=":", color="red", ax=ax)

        plt.xlabel("Timestep")
        plt.ylabel("Reserve (MW)")
        fig = plt.gcf()
    return fig


def plot_el_backup(sim_result, filename=None, showMargin=False, returnMargin=False):
    """plot reserve capacity vs device power output"""
    res = sim_result
    res_dev = res.el_backup
    dfP = res.device_flow.copy()
    carrier = "el"
    mask_carrier = dfP.index.get_level_values("carrier") == carrier
    mask_out = dfP.index.get_level_values("terminal") == "out"
    dfP.index = dfP.index.droplevel(level=("carrier", "terminal"))
    dfP = dfP[mask_carrier & mask_out].unstack(0)
    dfP = dfP[res_dev.columns]
    dfMargin = (res_dev - dfP).min(axis=1)
    if plotter == "plotly":
        fig = px.line()  # title="Online backup capacity (solid lines) vs device output (dotted lines)")
        colour = plotly.colors.DEFAULT_PLOTLY_COLORS
        k = 0
        for col in res_dev:
            fig.add_scatter(
                x=res_dev.index,
                y=res_dev[col],
                mode="lines",
                legendgroup=col,
                name="{} R_other".format(col),
                line_shape="hv",
                line=dict(color=colour[k]),
            )
            fig.add_scatter(
                x=dfP.index,
                y=dfP[col],
                legendgroup=col,
                name="{} P_out".format(col),
                line_shape="hv",
                line=dict(color=colour[k], dash="dot"),
            )
            k = k + 1
        if showMargin:
            fig.add_scatter(
                x=dfMargin.index,
                y=dfMargin,
                name="MARGIN",
                line=dict(color="black"),
                line_shape="hv",
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Power (MW)")
    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        ax = plt.gca()
        res_dev.plot(ax=ax, legend=True, alpha=1, linestyle="-")
        labels = list(res_dev.columns)
        if showMargin:
            dfMargin.plot(ax=ax, linestyle="-", linewidth=3, color="black", label="MARGIN")
            labels = labels + ["MARGIN"]
        plt.gca().set_prop_cycle(None)
        dfP.plot(ax=ax, linestyle=":", legend=False, alpha=1)
        plt.title("Online backup capacity (solid lines) vs device output (dotted lines)")
        ax.legend(labels, loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    if returnMargin:
        return dfMargin
    return fig


def recompute_elBackup(res, optimisation_model: pyo.Model):
    """Compute reserve
    should give the same as mc.compute_el_reserve"""
    optimiser = optimisation_model
    model = optimiser  # noqa
    all_devices = optimiser.all_devices

    # used capacity for all (relevant) devices:
    carrier = "el"
    devices_elin = optimiser.getDevicesInout(carrier_in=carrier)
    devices_elout = optimiser.getDevicesInout(carrier_out=carrier)
    dfP = res.device_flow
    mask_carrier = dfP.index.get_level_values("carrier") == carrier
    mask_out = dfP.index.get_level_values("terminal") == "out"
    mask_in = dfP.index.get_level_values("terminal") == "out"
    dfPin = dfP[mask_carrier & mask_in]
    dfPin.index = dfPin.index.droplevel(level=("carrier", "terminal"))
    dfPin = dfPin.unstack(0)
    dfPin = dfPin[devices_elin]
    dfPout = dfP[mask_carrier & mask_out]
    dfPout.index = dfPout.index.droplevel(level=("carrier", "terminal"))
    dfPout = dfPout.unstack(0)
    dfPout = dfPout[devices_elout]

    # reserve (unused) capacity by other devices:
    res_dev = pd.DataFrame(columns=dfPout.columns, index=dfPout.index)
    df_ison = res.device_is_on.unstack(0)
    for dev in devices_elout:
        otherdevs = [d for d in devices_elout if d != dev]
        cap_avail = 0
        for d in otherdevs:
            dev2 = all_devices[d].dev_data
            devmodel = dev2.model  # noqa
            maxValue = dev2.flow_max
            if dev2.profile is not None:
                extprofile = dev2.profile
                maxValue = maxValue * res.profiles_nowcast[extprofile]
            ison = 1
            if dev2.start_stop is not None:
                ison = df_ison[d]
            cap_avail += ison * maxValue
        otherdevs_in = [d for d in devices_elin if d != dev]  # noqa
        # TODO: include sheddable load
        res_dev[dev] = cap_avail - dfPout[otherdevs].sum(axis=1)
    return res_dev, dfPout


def plot_el_backup2(mc, filename=None):
    """plot reserve capacity vs device power output"""
    res_dev, dfP = recompute_elBackup(mc)
    plt.figure(figsize=(12, 4))
    ax = plt.gca()
    # default line zorder=2
    # res_dev.plot(ax=ax, linestyle="-", legend=True, alpha=0.5)
    res_dev.plot(ax=ax, drawstyle="steps-post", linestyle="-", legend=True, alpha=0.5)
    #    # critical is the largest load:
    #    c_critical = dfP.idxmax(axis=1)
    # critical is the smallest reserve margin:
    c_critical = (res_dev - dfP).idxmin(axis=1)  # noqa
    #    pd.Series(res_dev.lookup(res_dev.index,c_critical)).plot(ax=ax,
    #             linestyle='-',linewidth=2,label="CRITICAL",
    #             zorder=2.1,legend=True,color="black")
    (res_dev - dfP).min(axis=1).plot(ax=ax, linestyle="--", color="black", linewidth=2)

    # use the same colors
    plt.gca().set_prop_cycle(None)
    # dfP.plot(ax=ax, linestyle=":", legend=False, alpha=0.5)
    dfP.plot(ax=ax, drawstyle="steps-post", linestyle=":", legend=False, alpha=0.5)

    #    pd.Series(dfP.lookup(dfP.index,c_critical)).plot(ax=ax,
    #             linestyle=':',legend=False,linewidth=2,
    #             zorder=2.1,color="black")
    plt.title("Reserve capacity (solid line) vs device output (dotted line)")
    #    leg = ax.get_legend()
    #    leg.set_bbox_to_anchor((1.01,0))
    #    leg.set_frame_on(False)
    labels = list(dfP.columns) + ["MARGIN"]
    ax.legend(labels, loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
    #    plt.legend(dfP.columns,loc='lower left', bbox_to_anchor =(1.01,0),
    #               frameon=False)
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")
    return ax



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['oogeso',
 'oogeso.core',
 'oogeso.core.devices',
 'oogeso.core.networks',
 'oogeso.dto',
 'oogeso.io',
 'oogeso.plots',
 'oogeso.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Pyomo>=6.1.2,<7.0.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.21.4,<2.0.0',
 'pandas>=1.3.4,<2.0.0',
 'pyarrow>=6.0.1,<7.0.0',
 'pydot>=1.4.2,<2.0.0',
 'scipy>=1.7.2,<2.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'oogeso',
    'version': '0.4.0',
    'description': 'Offshore Oil and Gas Field Energy System Operational Optimisation (OOGESO)',
    'long_description': '<p>\n<a href="https://badge.fury.io/gh/oogeso%2Foogeso"><img src="https://badge.fury.io/gh/oogeso%2Foogeso.svg" alt="GitHub version" height="18"></a>\n<a href="https://github.com/oogeso/oogeso/actions/workflows/build.yml?query=workflow%3ACI"><img src="https://img.shields.io/github/workflow/status/oogeso/oogeso/CI" alt="Badge"></a>\n<a href="https://www.python.org/"><img src="https://img.shields.io/badge/python-3.7%20%7C%203.8%20%7C%203.9%20%7C%203.10-blue.svg" alt="Badge"></a>\n<a href="https://github.com/psf/black"><img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Badge"></a>\n<a href="https://lgtm.com/projects/g/oogeso/oogeso/alerts/"><img alt="Total alerts" src="https://img.shields.io/lgtm/alerts/g/oogeso/oogeso.svg?logo=lgtm&logoWidth=18"/></a>\n<a href="https://lgtm.com/projects/g/oogeso/oogeso/context:python"><img src="https://img.shields.io/lgtm/grade/python/g/oogeso/oogeso.svg?logo=lgtm&logoWidth=18" alt="Badge"></a>\n</p>\n<br/>\n\n# Offshore Oil and Gas Energy System Operational Optimisation Model (oogeso)\n\nPython module for modelling and analysing the energy system of offshore oil and gas fields, with renewable energy and storage integration.\n\nPart of the [Low Emission Centre](https://www.sintef.no/en/projects/lowemission-research-centre/) (SP5).\n\n## Getting started\nInstall latest Oogeso release from PyPi:\n```\npip install oogeso\n```\n\nin order to use the plotting functionality you will need to install plotting libraries:\n\n```\npip install matplotlib plotly seaborn\n```\n\n## User guide\nThe online user guide  gives more information about how to\nspecify input data and run a simulation case.\n\n*  [User guide](https://github.com/oogeso/oogeso/blob/master/userguide.md)\n\nThere is also a (not always up-to-date) manual with more information and explanations\nabout the modelling concepts and model elements:\n\n* [Manual (pdf)](https://github.com/oogeso/oogeso/blob/master/docs/oogeso_manual.pdf)\n\n## Examples\nCheck out the examples:\n\n* [Simple test case](https://github.com/oogeso/oogeso/blob/master/examples/test%20case2.ipynb)\n* [Test oil and gas platform](https://github.com/oogeso/oogeso/blob/master/examples/leogo_reference_platform.ipynb)\n\n## Local installation\nPrerequisite: \n- [Poetry](https://python-poetry.org/docs/#installation)\n- [Pre-commit](https://pre-commit.com/)\nClone or download the code and install it as a python package. I.e. navigate to the folder with the MANIFEST.in file and type:\n\n### Install dependencies\n1. `git clone git@github.com:oogeso/oogeso.git`\n2. `cd oogeso`\n3. `poetry install --no-root`  --no-root to not install the package itself, only the dependencies.\n4. `poetry shell`\n5. `poetry run pytests tests`\n\n### Local development in Docker\nAlternatively you can run and develop the code using docker and the Dockerfile in the root folder.\n\n### GitHub Actions Pipelines\n4 pipelines are defined.\n\n1. Build: Building and testing on multiple OS and python versions. Triggered on any push to GitHub.\n2. CBC-optimizer CI: Build and test oogeso with the CBC-solver and spesific cbc-tests.\n3. Release: Create release based on tags starting on v*.\n4. Publish: Publish the package to PyPi when a release is marked as published.\n\n## Contribute\nYou are welcome to contribute to the improvement of the code.\n\n* Use Issues to describe and track needed improvements and bug fixes\n* Use branches to avoid messing things up -- but don\'t veer too far away from the trunk (master branch)\n* Use [Pre-commit hooks](https://pre-commit.com/)\n\n### Contact\n\n[Harald G Svendsen](https://www.sintef.no/en/all-employees/employee/?empid=3414)  \nSINTEF Energy Research\n',
    'author': 'Harald Svendsen',
    'author_email': 'harald.svendsen@sintef.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/oogeso/oogeso',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



import logging
from typing import Union

import pyomo.environ as pyo

from oogeso.dto import NodeData

logger = logging.getLogger(__name__)


class NetworkNode:
    """Network node."""

    def __init__(self, node_data: NodeData):
        self.node_data = node_data
        self.id = node_data.id
        self.devices = {}
        self.devices_serial = {}  # devices with through-flow
        self.edges_from = {}
        self.edges_to = {}
        self.pressure_nominal = {}
        self.pressure_max_deviation = {}

    def get_pressure_nominal(self, carrier, term):
        if (carrier in self.pressure_nominal) and (term in self.pressure_nominal[carrier]):
            return self.pressure_nominal[carrier][term]
        return None

    def _set_pressure_parameter(self, carrier, term, parameter, value):
        if carrier in parameter:
            if term in parameter[carrier]:
                # pressure has already been set, so check consistency:
                p_exst = parameter[carrier][term]
                if p_exst != value:
                    raise Exception("Inconsistent pressure values for node {}:{} vs {}".format(self.id, p_exst, value))
            else:
                parameter[carrier][term] = value
        else:
            parameter[carrier] = {}
            parameter[carrier][term] = value

    def set_pressure_nominal(self, carrier, term, pressure):
        self._set_pressure_parameter(carrier, term, parameter=self.pressure_nominal, value=pressure)

    def set_pressure_maxdeviation(self, carrier, term, value):
        self._set_pressure_parameter(carrier, term, parameter=self.pressure_max_deviation, value=value)

    def add_device(self, device_id, device):
        # logger.debug("add_device: {},{}".format(self.id, device_id))
        self.devices[device_id] = device
        for carrier in device.serial:
            if carrier not in self.devices_serial:
                self.devices_serial[carrier] = {}
            self.devices_serial[carrier][device_id] = device

    def add_edge(self, edge, to_from: str):
        carrier = edge.edge_data.carrier
        edge_id = edge.id
        if to_from == "to":
            if carrier not in self.edges_to:
                self.edges_to[carrier] = {}
            self.edges_to[carrier][edge_id] = edge
        elif to_from == "from":
            if carrier not in self.edges_from:
                self.edges_from[carrier] = {}
            self.edges_from[carrier][edge_id] = edge

    def _rule_terminal_energy_balance(
        self, model: pyo.Model, carrier, terminal, t: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        r""" node energy balance (at in and out terminals)
        "in" terminal: flow into terminal is positive (Pinj>0)
        "out" terminal: flow out of terminal is positive (Pinj>0)

        distinguishing between the case (1) where node/carrier has a
        single terminal or (2) with an "in" and one "out" terminal
        (with device in series) that may have different properties
        (such as gas pressure)

        edge1 \                                     / edge3  # noqa
               \      devFlow_in    devFlow_out    /
                [term in]-->--device-->--[term out]
               /      \                       /    \
        edge2 /        \......termFlow......./      \ edge4


        """

        # Pinj = power injected into in terminal /out of out terminal
        Pinj = 0

        # Power in or out from connected devices:
        for dev_id, dev in self.devices.items():
            if (terminal == "in") and (carrier in dev.carrier_in):
                Pinj -= model.varDeviceFlow[dev_id, carrier, terminal, t]
            elif (terminal == "out") and (carrier in dev.carrier_out):
                Pinj -= model.varDeviceFlow[dev_id, carrier, terminal, t]

        # If no device is connected between in and out terminal for a given
        # energy carrier, connect the terminals (treat as one):
        if carrier not in self.devices_serial:
            Pinj -= model.varTerminalFlow[self.id, carrier, t]

        # edges:
        # losses are zero for all but "el" edges
        # "in": edgeFlow12 - loss12 - edgeFlow21 = edgeFlow - loss12
        # "out": edgeFlow12 - (edgeFlow21-loss21) = edgeFlow + loss21
        if (terminal == "in") and (carrier in self.edges_to):
            for edge_id, edge in self.edges_to[carrier].items():
                # power into node from edge
                if edge.has_loss():
                    Pinj += model.varEdgeFlow[edge_id, t] - model.varEdgeLoss12[edge_id, t]
                else:
                    Pinj += model.varEdgeFlow[edge_id, t]
        elif (terminal == "out") and (carrier in self.edges_from):
            for edge_id, edge in self.edges_from[carrier].items():
                # power out of node into edge
                if edge.has_loss():
                    Pinj += model.varEdgeFlow[edge_id, t] + model.varEdgeLoss21[edge_id, t]
                else:
                    Pinj += model.varEdgeFlow[edge_id, t]

        # if (carrier,node) in model.paramNodeEdgesTo and (terminal=='in'):
        #     for edg in model.paramNodeEdgesTo[(carrier,node)]:
        #         # power into node from edge
        #         Pinj += (model.varEdgeFlow[edg,t])
        # elif (carrier,node) in model.paramNodeEdgesFrom and (terminal=='out'):
        #     for edg in model.paramNodeEdgesFrom[(carrier,node)]:
        #         # power out of node into edge
        #         Pinj += (model.varEdgeFlow[edg,t])

        expr = Pinj == 0
        if (type(expr) is bool) and (expr is True):
            expr = pyo.Constraint.Skip
        return expr

    def _rule_pressure_at_node(self, model: pyo.Model, carrier, t) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        node = self.id
        if carrier in ["el", "heat"]:
            return pyo.Constraint.Skip
        elif carrier in self.devices_serial:
            # pressure in and out are related via device equations for
            # device connected between in and out terminals. So no
            # extra constraint required
            return pyo.Constraint.Skip
        else:
            # single terminal. (pressure out=pressure in)
            expr = model.varPressure[(node, carrier, "out", t)] == model.varPressure[(node, carrier, "in", t)]
            return expr

    def _rule_pressure_bounds(
        self, model: pyo.Model, term, carrier, t: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        node = self.id
        nominal_pressure = self.pressure_nominal
        max_dev = None  # default is no constraint
        if carrier in nominal_pressure:
            if term in nominal_pressure[carrier]:
                nom_p = nominal_pressure[carrier][term]
                if nom_p is not None:
                    if (carrier in self.pressure_max_deviation) and (term in self.pressure_max_deviation[carrier]):
                        max_dev = self.pressure_max_deviation[carrier][term]
                        if t == 0:
                            logger.debug(
                                "Using individual pressure limit for: {}, {}, {}, {}".format(
                                    node, carrier, term, max_dev
                                )
                            )
                    else:
                        # Using globally set pressure deviation limit
                        max_dev = model.paramMaxPressureDeviation
            else:
                nom_p = 0
                # raise ValueError("nom_p has no value")  # Fixme: How do we want to handle this case? nom_p unresolved.
        else:
            nom_p = 0
            # raise ValueError("nom_p has no value")  # Fixme: How do we want to handle this case? nom_p is unresolved.
        if (max_dev is None) or (max_dev == -1):
            return pyo.Constraint.Skip
        lower_bound = nom_p * (1 - max_dev)
        upper_bound = nom_p * (1 + max_dev)
        expr = pyo.inequality(lower_bound, model.varPressure[(node, carrier, term, t)], upper_bound)
        return expr

    def define_constraints(self, pyomo_model: pyo.Model):
        """Returns the set of constraints for the node."""
        model = pyomo_model

        constr_terminal_energy_balance = pyo.Constraint(
            model.setCarrier,
            ["in", "out"],
            model.setHorizon,
            rule=self._rule_terminal_energy_balance,
        )
        setattr(
            model,
            f"constrN_{self.id}_energybalance",
            constr_terminal_energy_balance,
        )

        constr_pressure_at_node = pyo.Constraint(model.setCarrier, model.setHorizon, rule=self._rule_pressure_at_node)
        setattr(
            model,
            f"constrN_{self.id}_nodepressure",
            constr_pressure_at_node,
        )

        constr_pressure_bounds = pyo.Constraint(
            model.setTerminal,
            model.setCarrier,
            model.setHorizon,
            rule=self._rule_pressure_bounds,
        )
        setattr(
            model,
            f"constrN_{self.id}_pressurebound",
            constr_pressure_bounds,
        )

    def is_non_trivial(self, carrier):
        """returns True if edges or devices are connected for given carrier"""
        # check if any edges are connected on this carrier
        if carrier in self.edges_from:
            return True
        if carrier in self.edges_to:
            return True
        for dev_obj in self.devices.values():
            # check if any devices are connected on this carrier
            if carrier in dev_obj.carrier_in:
                return True
            if carrier in dev_obj.carrier_out:
                return True
        # nothing connected:
        return False



import json
import logging
from dataclasses import asdict, is_dataclass
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from oogeso import dto
from oogeso.utils.util import get_class_from_dto

logger = logging.getLogger(__name__)


class DataclassJSONEncoder(json.JSONEncoder):
    """Serialize (save to file)"""

    def default(self, obj: Any):
        if is_dataclass(obj=obj):
            dct = asdict(obj=obj)
            return dct
        if isinstance(obj, pd.Series):
            dct = obj.reset_index().to_json(orient="records")
            return dct
        if isinstance(obj, pd.DataFrame):
            dct = obj.reset_index().to_json(orient="records")
            return dct
        return super().default(obj)


class DataclassJSONDecoder(json.JSONDecoder):
    """Deserialize (read from file)"""

    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def _new_carrier(self, dct: Dict[str, str]):
        model = dct["id"]
        carrier_class_str = f"Carrier{model.capitalize()}Data"
        carrier_class = get_class_from_dto(class_str=carrier_class_str)
        return carrier_class(**dct)

    def _new_node(self, dct: Dict[str, str]) -> dto.NodeData:
        return dto.NodeData(**dct)

    def _new_edge(self, dct: Dict[str, str]) -> dto.EdgeData:
        carrier = dct.pop("carrier")  # gets and deletes model from dictionary
        edge_class_str = f"Edge{carrier.capitalize()}Data"
        edge_class = get_class_from_dto(class_str=edge_class_str)
        return edge_class(**dct)

    def _new_device(self, dct: Dict[str, Union[str, object]]) -> dto.DeviceData:
        logger.debug(dct)
        model = dct.pop("model")  # gets and deletes model from dictionary
        start_stop: Optional[Dict] = dct.pop("start_stop", None)
        if start_stop is not None:
            startstop_obj = dto.StartStopData(**start_stop)
            dct["start_stop"] = startstop_obj
        dev_class_str = f"Device{model.capitalize()}Data"
        dev_class = get_class_from_dto(class_str=dev_class_str)
        return dev_class(**dct)

    @staticmethod
    def _new_profile(dct: Dict[str, Union[str, List[float]]]) -> dto.TimeSeriesData:
        name = dct["id"]
        data = dct["data"]
        data_nowcast = None
        if "data_nowcast" in dct:
            data_nowcast = dct["data_nowcast"]
        return dto.TimeSeriesData(name, data, data_nowcast)

    def object_hook(self, dct):  # pylint: disable=E0202
        carriers = []
        nodes = []
        edges = []
        # devs = {}
        devs = []
        profiles = []
        # profiles_nowcast = {}
        if "nodes" in dct:
            # Top level
            d_params = dct["parameters"]
            params = dto.OptimisationParametersData(**d_params)
            for n in dct["carriers"]:
                carriers.append(self._new_carrier(n))
            for n in dct["nodes"]:
                nodes.append(self._new_node(n))
            for n in dct["edges"]:
                edges.append(self._new_edge(n))
            for n in dct["devices"]:
                devs.append(self._new_device(n))
            if "profiles" in dct:
                for n in dct["profiles"]:
                    profiles.append(self._new_profile(n))
                    # profiles[i] = self._new_profile(n)
            # if "profiles_nowcast" in dct:
            #    for i, n in dct["profiles_nowcast"].items():
            #        profiles_nowcast[i] = self._new_profile(n)
            energy_system_data = dto.EnergySystemData(
                carriers=carriers,
                nodes=nodes,
                edges=edges,
                devices=devs,
                parameters=params,
                profiles=profiles,
                # profiles_nowcast=profiles_nowcast,
            )
            return energy_system_data
        return dct

    def default(self, obj: Any):
        if isinstance(obj, dto.EnergySystemData):
            return dto.EnergySystemData(obj=obj)
        return json.JSONEncoder.default(self, obj)


def serialize_oogeso_data(energy_system_data: dto.EnergySystemData):
    json_string = json.dumps(energy_system_data, cls=DataclassJSONEncoder, indent=2)
    return json_string


def deserialize_oogeso_data(json_data):
    energy_system_data = json.loads(json_data, cls=DataclassJSONDecoder)
    return energy_system_data


# Deserialize result object
class OogesoResultJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):  # pylint: disable=E0202
        res_dfs = {}
        # profiles_nowcast = {}
        if "device_flow" in dct:
            # Top level
            for k, val in dct.items():
                print(k)
                if val is None:
                    res_dfs[k] = None
                else:
                    val2 = json.loads(val)
                    df = pd.DataFrame.from_dict(val2)
                    if df.empty:
                        df = None
                    elif k in ["profiles_forecast", "profiles_nowcast"]:
                        df = df.set_index("index")
                    else:
                        # change to multi-index pandas Series
                        index_names = ["device", "time", "carrier", "terminal", "edge", "node"]
                        multiind = [c for c in df.columns if c in index_names]
                        # print(df.columns)
                        # if multiind == []:
                        #    print(val, type(val))
                        df = df.set_index(multiind)
                        # print(df.columns)
                        df = df.iloc[:, 0]
                    res_dfs[k] = df
            result_data = dto.SimulationResult(**res_dfs)
            return result_data
        return dct

    def default(self, obj: Any):
        if isinstance(obj, dto.SimulationResult):
            return dto.SimulationResult(obj=obj)
        return json.JSONEncoder.default(self, obj)


def deserialize_oogeso_results(json_data):
    result_data = json.loads(json_data, cls=OogesoResultJSONDecoder)
    return result_data


if __name__ == "__main__":
    # Todo: Move to examples and/or tests.
    print("Serializing example data and saving to file (examples/energysystem.json)")

    # Example:
    energy_system = dto.EnergySystemData(
        carriers=[
            dto.CarrierElData(
                id="el",
                reference_node="node1",
                el_reserve_margin=-1,
            ),
            dto.CarrierHeatData("heat"),
            dto.CarrierGasData(
                "gas",
                co2_content=0.4,
                G_gravity=0.6,
                Pb_basepressure_MPa=100,
                R_individual_gas_constant=9,
                Tb_basetemp_K=300,
                Z_compressibility=0.9,
                energy_value=40,
                k_heat_capacity_ratio=0.7,
                rho_density=0.6,
            ),
        ],
        nodes=[dto.NodeData("node1"), dto.NodeData("node2")],
        edges=[
            dto.EdgeElData(
                id="edge1",
                node_from="node1",
                node_to="node2",
                flow_max=500,
                reactance=1,
                resistance=1,
                voltage=33,
                length_km=10,
            ),
            dto.EdgeElData(
                id="edge2",
                node_from="node2",
                node_to="node1",
                flow_max=50,
                voltage=33,
                length_km=10,
                # reactance=1,
                # resistance=1,
                power_loss_function=([0, 1], [0, 0.02]),
            ),
        ],
        devices=[
            dto.DeviceSourceElData(id="elsource", node_id="node1", flow_max=12),
            dto.DevicePowerSourceData(id="gt1", node_id="node2", flow_max=30, profile="profile1"),
            dto.DeviceSinkElData(id="demand", node_id="node2", flow_min=4, profile="profile1"),
        ],
        parameters=dto.OptimisationParametersData(
            time_delta_minutes=30,
            planning_horizon=12,
            optimisation_timesteps=6,
            forecast_timesteps=6,
            time_reserve_minutes=30,
            max_pressure_deviation=-1,
            co2_tax=30,
            objective="exportRevenue",
        ),
        profiles=[dto.TimeSeriesData(id="profile1", data=[12, 10, 21])],
    )

    # serialized = serialize_oogeso_data(energy_system)
    # print(serialized)

    with open("examples/energysystem.json", "w") as outfile:
        json.dump(energy_system, fp=outfile, cls=DataclassJSONEncoder, indent=2)
